(**************************************************************************)
(*                                                                        *)
(*                     TRX: A Web Transaction Manager                     *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Defining back-end functions 
   for maintaining memory of transactions and processing input/output *) 
open TyTr;
open App;

(* The user table file has format [(current session number, hash_table)]
   where [hash_table] keeps triples [(user * session * transaction)] *)

(* Functor taking as parameter the application module conforming to [APP] *)
module Back = 
  functor (A : APP) -> struct

exception BackEndError of string;

(* -- retrieving session from user table file *)
(* [getSession : string -> session -> session ] *)
value getSession fn w = 
  let sfin = open_in fn in
  let (w',tab) = input_value sfin in
  let _ = close_in sfin in
  let w' = if w = 0 then (w'+1) (* new session *)
           else if w > 0 && w <= w' then w (* old session *)
                else raise (BackEndError "Invalid session") in
  let sfout = open_out fn in
  let _ = output_value sfout (w',tab) in
  let _ = close_out sfout in
  w';

(* -- retrieving memory from user table file *)
(* [ getEntry : string -> (user * session * transaction) -> App.stt ] *) 
value getEntry fn (c,w,n) = 
  let sfin = open_in fn in
  let (w',tab) = input_value sfin in
  let _ = close_in sfin in
  let entry:(A.stt) = try Hashtbl.find tab (c,w,n)
  with [ _ -> if n = 1 then (* 1st transaction of a new session *)
		let _ = Hashtbl.add tab (c,w,n) A.initstt in
                let sfout = open_out fn in
                let _ = output_value sfout (w',tab) in
                let _ = close_out sfout in A.initstt
              else raise (BackEndError "Invalid transaction number")
       ] in entry;

(* -- updating memory to user table file *)
(* [ updateEntry : string -> (user * session * transaction) -> App.stt -> unit ] *)
value updateEntry fn (c,w,n) mem = 
  let sfin = open_in fn in
  let (w',tab) = input_value sfin in
  let _ = close_in sfin in
  let _ = Hashtbl.replace tab (c,w,n) mem in
  let sfout = open_out fn in
  let _ = output_value sfout (w',tab) in
  close_out sfout;

(* -- processing input for [NEXT] transaction (calls [updateEntry]) *)
(* [ processInput : string -> App.stt -> (idstamp * arglist) -> string ] *)
value processInput fn mem ((s,c,w,n),alist) =
  let vce = (mem,alist) in
  let vb = eng A.cof (s,c,w,n) vce in
  let vb' = match vb with 
    [ (res,olist,mem') -> do
      { updateEntry fn (c,w,n) mem'
      ; (res,olist) 
      } ] in
  bkend A.prf vb';

(* -- ignoring input and recalling [JUMP] transaction *)
(*	[ ignoreInput : App.stt -> (idstamp * arglist) -> string ] *)
value ignoreInput mem (idtup,_) =
  let vb = trv A.jmf idtup mem in bkend A.prf vb;	

(* -- setting up client for transaction manager daemon,
  establishes socket to receive [((idstamp * arglist) * filename)],
  processes the transaction and passes back [string] to the daemon 
  (values are marshalled using functions [output_value] and [input_value]) *)
value client_up ip port =
  let sock = (Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0) in 
  let ser_addr = Unix.ADDR_INET (Unix.inet_addr_of_string ip) port in
  let success = ref False in
  let _ = while (success.val = False) do (* may loop if connection fails *) 
  	{ try do { Unix.connect sock ser_addr 
                 ; success.val := True
                 } 
	  with [ _ -> () ]
        } in
  let sock_in = Unix.in_channel_of_descr sock
  and sock_out = Unix.out_channel_of_descr sock in
  let (vtm,callt) = input_value sock_in in
  let out_str = match vtm with 
   [ ((s,c,w,n),alist) -> if n = 1 then 
    try let w = getSession callt w 
        in ((hdtrx s c w n A.initsub) ^ (A.initstr c) ^ (tltrx s c w n)) 
    with [ BackEndError err -> 
           ((hdtrx s c w n False) ^ (prerr err) ^ (tltrx s c w (n+1))) 
         ]               else 
    match (List.assoc "jop" alist) with 
     [ "next" -> try let mem = getEntry callt (c,w,n-1) in
           ((hdtrx s c w n False) ^ try processInput callt mem vtm
                                    with [ A.AppError err -> prerr err ] ^
            (tltrx s c w (n+1)))
	         with [ BackEndError err -> 
            ((hdtrx s c w n False) ^ (prerr err) ^ (tltrx s c w (n+1))) ]
     | "jump" ->
  	try let tn = 
          try (int_of_string (List.assoc "trxno" alist)) 
          with [ _ -> raise (BackEndError "Invalid transaction number") ] 
            in let mem = getEntry callt (c,w,tn) in
            ((hdtrx s c w tn False) ^ (ignoreInput mem vtm) ^ (tltrx s c w (tn+1))) 
	with [ BackEndError err -> 
            ((hdtrx s c w n False) ^ (prerr err) ^ (tltrx s c w (n+1))) ]
     | "quit" -> deftrx
     | _ -> raise (BackEndError "Invalid option") 
     ] 
   ] 
  in do  
    { output_value sock_out out_str
    ; flush sock_out
    ; Unix.close sock
    ; flush stderr
    };

end;
