(**************************************************************************)
(*                                                                        *)
(*                  STRED: A Distributed Structure Editor                 *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

open TyEd;
open Table;

(* -- defining an ML type
	for the restricted XML generated by IoXML *) 
type xmlEd = 
	[ NODE of string and list xmlEd
	| BRANCH of string and xmlEd
	| SING of string
	| TUPLE of list xmlEd
	| LIST of list xmlEd
	| OPTION of option xmlEd
	| TI of xmlEd
	| LI of xmlEd
	| CI of xmlEd
	| STRING of string
	| EMPTY
	| FOCUS
	| ETC ];

(* -- getting rid of element delimiters,
	in syntax of IoXML *) 
value rec simp v = 
	match v with
	[ NODE c fl -> NODE c (List.map simp fl)
	| BRANCH c f -> NODE c [simp f]
	| SING "o" -> EMPTY
	| SING c -> NODE c []
	| LIST fl -> LIST (List.map simp fl)
	| TUPLE fl -> TUPLE (List.map simp fl)
	| OPTION (Some f) -> OPTION (Some (simp f))
	| CI f -> simp f
	| TI f -> simp f
	| LI f -> simp f
	| x -> x ];

(* -- prunes xml-tree to a give depth *)
value rec prunetree delta v = 
  if delta = 0 then ETC 
  else match v with
    [ NODE c [_::_] when delta = 1 -> NODE c [ETC]
    | LIST [_::_] when delta = 1 -> LIST [ETC]
    | TUPLE _ when delta = 1 -> TUPLE [ETC]
    | OPTION (Some f) when delta = 1 -> OPTION (Some ETC)
    | NODE c fl -> NODE c (List.map (prunetree (delta - 1)) fl)
    | LIST fl -> LIST (List.map (prunetree (delta - 1)) fl)
    | TUPLE fl -> TUPLE (List.map (prunetree (delta - 1)) fl)
    | OPTION (Some f) -> OPTION (Some (prunetree (delta - 1) f))
    | _ -> v 
    ];

value rec indent i = 
  if i=0 then "" else " " ^ (indent (i - 1));
	
exception XmlEdError of string;
		
(* -- defining zipper type 
for (simlplified) IoXML-generated XML *)

type zipxmlEd = 
  [ Out
  | Top
  (* type * contructor, left, up, right *) 
  | In_NODE of tyEd and string and list xmlEd and zipxmlEd and list xmlEd
  (* left, up, right *)
  | In_TUPLE of tyEd and list xmlEd and zipxmlEd and list xmlEd
  | In_LIST of tyEd and list xmlEd and zipxmlEd and list xmlEd
  (* up *)
  | In_OPTION of tyEd and zipxmlEd 
  ];

(* -- prunes zipper *)
value rec prunezip delta z s = 
  if z = Top then (s,True,delta) 
  else match delta with
    [ 0 -> (s,False,0)
    | _ -> let (l,r) = match z with
      [ In_NODE _ _ l _ r -> (l,r)
      | In_TUPLE _ l _ r -> (l,r)
      | In_LIST _ l _ r -> (l,r)
      | In_OPTION _ _ -> ([],[]) 
      | _ -> raise (XmlEdError "Unprunable zipper") 
      ] in let g lst' = fst (List.fold_left 
             (fun (lst,d) x -> match d with 
               [ 0 -> ([ETC::lst],-1)
               | -1 -> (lst,-1) 
               | _ -> ([(prunetree d x)::lst],d - 1) 
               ])
			([],delta) lst') in
           let l'= List.rev (g l)
           and r'= List.rev (g r) in
           match z with
		[ In_NODE _ c _ z' _ -> prunezip (delta - 1) z' 
					(NODE c (List.rev_append l' [s::r'])) 
		| In_TUPLE _ _ z' _ -> prunezip (delta - 1) z'
					(TUPLE (List.rev_append l' [s::r']))
		| In_LIST _ _ z' _ -> prunezip (delta - 1) z'
					(LIST (List.rev_append l' [s::r']))
		| In_OPTION _ z'-> prunezip (delta - 1) z'
					(OPTION (Some s)) 
		| _ -> raise (XmlEdError "Unprunable zipper") 
                ] 
    ];

(* -- typechecker *)
value rec tycheck hash t s = let t = 
  match t with
   [ Ty_user tn -> let (_,(nt,_)) = follow hash tn in nt
   | _ -> t
   ] in let _ = 
     match s with
      [ NODE c fl -> match t with 
        [ Ty_user tn -> let e = Hashtbl.find hash tn in
                        match List.mem c (get_clist e) with
                           [ True -> let tl = get_flist e c in
                                     ignore (List.map2 (tycheck hash) tl fl)
                           | _ -> raise (XmlEdError "Invalid constructor") 
                           ] 
        | _ -> raise (XmlEdError "Wrong type association") 
        ]
      | LIST fl -> match t with 
        [ Ty_LIST t' -> ignore (List.map (tycheck hash t') fl) 
	| _ -> raise (XmlEdError "Wrong type association") 
        ]
      | OPTION f -> match t with 
        [ Ty_OPTION t' -> match f with 
          [ Some s' -> ignore (tycheck hash t' s')
          | None -> () 
          ] 
        | _ -> raise (XmlEdError "Wrong type association") 
        ]
      | TUPLE fl -> match t with 
          [ Ty_TUPLE tl -> ignore (List.map2 (tycheck hash) tl fl) 
          | _ -> raise (XmlEdError "Wrong type association") 
          ]
      | STRING x -> match t with 
          [ Ty_INT -> ignore (int_of_string x)
          | Ty_STRING -> () 
          | _ -> raise (XmlEdError "Wrong type association") 
          ]
      | EMPTY -> ()
      | _ -> raise (XmlEdError "Not type-checkable xml") 
      ]
        in t;

