(**************************************************************************)
(*                                                                        *)
(*                  STRED: A Distributed Structure Editor                 *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Stripped version of ioXML code,
   XML printing replaced by conversion to XML type *)

open TyEd;
open Table;
open XmlEd;

exception XmlppError of string;

type ast =
  [ Tag of (int * int) and string and list ast
  | Str of (int * int) and string ]
;

exception ExcLoc of (int * int) and exn;

value raise_loc loc exc =
  match exc with
  [ ExcLoc _ _ -> raise exc
  | _ -> raise (ExcLoc loc exc) ]
;

value error_txt loc str = raise_loc loc (Stream.Error str);
value error loc str = error_txt loc (str ^ " expected");

(* -- buffer machinery *)
value buff = Buffer.create 80;
value store len c =
  do {
    if len = 0 then Buffer.clear buff else ();
    Buffer.add_char buff c;
    Buffer.length buff
  }
;
value get_buff len = if len = 0 then "" else Buffer.contents buff;

value valch x = Char.code x - Char.code '0';

value rec backslash s i =
  if i = String.length s then raise Not_found
  else
    match s.[i] with
    [ 'n' -> ('\n', i + 1)
    | 'r' -> ('\r', i + 1)
    | 't' -> ('\t', i + 1)
    | 'b' -> ('\b', i + 1)
    | '\\' -> ('\\', i + 1)
    | '0'..'9' as c -> backslash1 (valch c) s (i + 1)
    | _ -> raise Not_found ]
and backslash1 cod s i =
  if i = String.length s then ('\\', i - 1)
  else
    match s.[i] with
    [ '0'..'9' as c -> backslash2 (10 * cod + valch c) s (i + 1)
    | _ -> ('\\', i - 1) ]
and backslash2 cod s i =
  if i = String.length s then ('\\', i - 2)
  else
    match s.[i] with
    [ '0'..'9' as c -> (Char.chr (10 * cod + valch c), i + 1)
    | _ -> ('\\', i - 2) ]
;

value rec skip_indent s i =
  if i = String.length s then i
  else
    match s.[i] with
    [ ' ' | '\t' -> skip_indent s (i + 1)
    | _ -> i ]
;

value skip_opt_linefeed s i =
  if i = String.length s then i else if s.[i] = '\010' then i + 1 else i
;

value token_eval_char s =
  if String.length s = 1 then s.[0]
  else if String.length s = 0 then failwith "invalid char token"
  else if s.[0] = '\\' then
    if String.length s = 2 && s.[1] = ''' then '''
    else
      try
        let (c, i) = backslash s 1 in
        if i = String.length s then c else raise Not_found
      with
      [ Not_found -> failwith "invalid char token" ]
  else failwith "invalid char token"
;

value token_eval_string s =
  loop 0 0 where rec loop len i =
    if i = String.length s then get_buff len
    else
      let (len, i) =
        if s.[i] = '\\' then
          let i = i + 1 in
          if i = String.length s then failwith "invalid string token"
          else if s.[i] = '"' then (store len '"', i + 1)
          else
            match s.[i] with
            [ '\010' -> (len, skip_indent s (i + 1))
            | '\013' -> (len, skip_indent s (skip_opt_linefeed s (i + 1)))
            | c ->
                try
                  let (c, i) = backslash s i in
                  (store len c, i)
                with
                [ Not_found -> (store (store len '\\') c, i + 1) ] ]
        else (store len s.[i], i + 1)
      in
      loop len i
;

(* -- parsing from stream to ast *)
type loc = (int * int);

type token =
  [ Tequal of loc
  | Teoi of loc
  | Tgreater of loc
  | Tident of loc and string
  | Tint of loc and string
  | Tless of loc
  | Tlessslash of loc
  | Tslash of loc
  | Tstring of loc and string ]
;

value rec ident len =
  parser
  [ [: `(('a'..'z' | 'A'..'Z' | '_' | '0'..'9' | ''') as c); s :] ->
      ident (store len c) s
  | [: :] -> get_buff len ]
;

value rec number len =
  parser
  [ [: `('0'..'9' as c); s :] -> number (store len c) s
  | [: :] -> get_buff len ]
;

value rec string len =
  parser
  [ [: `'"' :] -> get_buff len
  | [: `'\\'; `c; s :] -> string (store (store len '\\') c) s
  | [: `c; s :] -> string (store len c) s ]
;

value less bp =
  parser
  [ [: `'/' :] ep -> Tlessslash (bp, ep)
  | [: :] ep -> Tless (bp, ep) ]
;

value rec lexer_func =
  parser bp
  [ [: `' ' | '\t' | '\r' | '\n'; s :] -> lexer_func s
  | [: `'<'; t = less bp :] -> t
  | [: `'"'; s = string 0 :] ep -> Tstring (bp, ep) s
  | [: `('0'..'9' | '-' as c); s = number (store 0 c) :] ep -> Tint (bp, ep) s
  | [: `('a'..'z' | 'A'..'Z' as c); s = ident (store 0 c) :] ep ->
      Tident (bp, ep) s
  | [: `'=' :] ep -> Tequal (bp, ep)
  | [: `'/' :] ep -> Tslash (bp, ep)
  | [: `'>' :] ep -> Tgreater (bp, ep)
  | [: :] ep -> Teoi (bp, ep) ]
;

value rec xml_list rev_list =
  parser
  [ [: x = xml; s :] -> xml_list [x :: rev_list] s
  | [: :] -> List.rev rev_list ]
and xml =
  parser
  [ [: `Tless (bp, _); `Tident _ bt  ? "identifier expected";
       attr = attr_list [];
       t =
         parser
         [ [: `Tslash _; `Tgreater (_, ep) ? "'>' expected" :] ->
             Tag (bp, ep) bt []
         | [: `Tgreater (_, ep); el = xml_list [];
               (et, loc1) = end_tag ? "end tag expected" :] ->
             if et <> bt then
               raise_loc loc1
                 (Stream.Error ("</" ^ bt ^ "> expected"))
             else Tag (bp, ep) bt el ] :] -> t
  | [: (s, loc) = string :] -> Str loc s ]
and end_tag =
  parser
  [ [: `Tlessslash (bp, _); `Tident _ i ? "identifier expected";
       `Tgreater (_, ep) ? "'>' expected" :] ->
      (i, (bp, ep)) ]
and attr_list rev_list =
  parser
  [ [: x = attribute; s :] -> attr_list [x :: rev_list] s
  | [: :] -> List.rev rev_list ]
and attribute =
  parser
  [ [: `Tident _ k; `Tequal _ ? "'=' expected";
       (v, _) = string ? "string or integer expected" :] -> (k, v) ]
and string =
  parser
  [ [: `Tstring loc s :] -> (s, loc)
  | [: `Tint loc s :] -> (s, loc) ]
;

value get_loc cstrm strm =
  match Stream.peek strm with
  [ Some x ->
      match x with
      [ Tequal loc
      | Teoi loc
      | Tgreater loc
      | Tident loc _
      | Tint loc _
      | Tless loc
      | Tlessslash loc
      | Tslash loc
      | Tstring loc _ -> loc ]
  | _ -> let bp = Stream.count strm in (bp, bp + 1) ]
;

value parse_xml cstrm =
  let strm = Stream.from (fun _ -> Some (lexer_func cstrm)) in
  try xml strm with e -> raise_loc (get_loc cstrm strm) e
;

value parse_xml_list cstrm =
  let strm = Stream.from (fun _ -> Some (lexer_func cstrm)) in
  let r =
    try xml_list [] strm with e ->
      raise_loc (get_loc cstrm strm) e
  in
  match strm with parser
  [ [: `Teoi _ :] -> r
  | [: :] ->
      raise_loc (get_loc cstrm strm)
        (Stream.Error ("End of input expected")) ]
;

(* -- parsing ast *)
value rec xmlEd_of_ast = fun
	[ Tag _ "ci" [a] -> CI (xmlEd_of_ast a)
	| Tag _ "li" [a] -> LI (xmlEd_of_ast a)
	| Tag _ "ti" [a] -> TI (xmlEd_of_ast a)
	| Tag _ "tuple" al -> TUPLE (List.map xmlEd_of_ast al)
	| Tag _ "list" al -> LIST (List.map xmlEd_of_ast al)
	| Tag _ "Some" [a] -> OPTION (Some (xmlEd_of_ast a))
	| Tag _ "None" [] -> OPTION (None)
	| Tag _ lab [] -> SING lab
	| Tag _ lab [a] -> BRANCH lab (xmlEd_of_ast a)
	| Tag _ lab al -> NODE lab (List.map xmlEd_of_ast al)
	| Str _ s -> STRING (token_eval_string s) ];
	
value pp_parse strm = simp (xmlEd_of_ast (parse_xml strm));

value rec prxml v t hash i = 
	let rec indent j = 
		match j with
		[ 0 -> ""
		| _ -> " "^ (indent (j - 1)) ] in
	match v with
	[ NODE c fl -> 
		match t with [ Ty_user tn ->
		let e = Hashtbl.find hash tn in
		let tl = get_flist e c in
		match fl with
		[ [] -> 
			(indent i)^"<"^c^"/>\n"
		| [f] -> 
			(indent i)^"<"^c^">\n"^
			(prxml f (List.hd tl) hash (i+1))^
			(indent i)^"</"^c^">\n"
		| _ -> 
			(indent i)^"<"^c^">\n"^
			(fst (List.fold_left 
				(let f (s,tl) = 
				fun [ x -> 
					(s^
					(indent (i+1))^"<ci>\n"^
					(prxml x (List.hd tl) hash (i+2))^
					(indent (i+1))^"</ci>\n",
						(List.tl tl)) ] in f)
			("",tl) fl))^(indent (i+1))^"</"^c^">\n" ] 
			| _ -> raise (XmlppError "Wrong type association") ]

	| LIST fl -> 
		match t with [ Ty_LIST t' -> 
		(indent i)^"<list>\n"^
		(List.fold_left 
			(let f s = 
			fun [ x -> 
				s^ 
				(indent (i+1))^"<li>\n"^
				(prxml x t' hash (i+2))^
				(indent (i+1))^"</li>\n" ] in f)
		"" fl)^
		(indent i)^"</list>\n" 
		| _ -> raise (XmlppError "Wrong type association") ]

	| TUPLE fl -> 
		match t with [ Ty_TUPLE tl ->
		(indent i)^"<tuple>\n"^
		(fst (List.fold_left 
			(let f (s,tl) = 
			fun [ x -> 
				(s^ 
				(indent (i+1))^"<ti>\n"^
				(prxml x (List.hd tl) hash (i+2))^
				(indent (i+1))^"</ti>\n",(List.tl tl)) ]
				in f) 
		("",tl) fl))^
		(indent i)^"</tuple>\n" 
		| _ -> raise (XmlppError "Wrong type association") ]

	| OPTION None -> 
		(indent i)^"<None/>\n"

	| OPTION (Some f) -> 
		match t with [ Ty_OPTION t' ->
		(indent i)^"<Some>\n"^
		(prxml f t' hash (i+1))^
		(indent i)^"</Some>\n" 
		| _ -> raise (XmlppError "Wrong type association") ]

	| STRING x -> 
		(indent i)^
		match t with 
		[ Ty_INT -> x
		| Ty_STRING -> "\""^x^"\"" 
		| _ -> raise (XmlppError "Wrong type association") ]^"\n" 
		
	| _ -> raise (XmlppError "Not printable xml") ];

value pp_print v t hash i = prxml v t hash i;

